//Title: Device Manager Class
//Platform: BeOS 5
//Version: 2006.01.18
//Description:
//	A class for working with physical hardware devices (like VGA displays)
//  Basically a factory class for the other Device classes.

//Copyright (C) 2002-2006 Kevin H. Patterson

//VGA Device Output consumer node, 18 january 2006, alpha 3 release.


#include <dirent.h>
#include <string.h>

//#include <iostream.h>

#include "DeviceManager.h"
#include "Log.h"

DeviceManager::DeviceManager()
{
}

DeviceManager::~DeviceManager()
{
}

GraphicsDevice * DeviceManager::GetGraphicsDevice(const char * DeviceName)
{
	GraphicsDevice * tGraphicsDevice(new GraphicsDevice(DeviceName));
	if (tGraphicsDevice != NULL)
	{
		if (tGraphicsDevice->InitCheck())
		{
			//this graphics card is supported
			return tGraphicsDevice;
		}
		else
		{
			//device not found or not supported 
			delete tGraphicsDevice;
			return NULL;
		}
	}
	else
		return NULL;
}

GraphicsDevice * DeviceManager::GetGraphicsDevice(uint32 DeviceIndex)
{
	GraphicsDevice * tGraphicsDevice(NULL);

	uint32 DeviceCount(0);
	const char * DevicePath = "/dev/graphics";
	DIR * DeviceDirectory;
	DeviceDirectory = opendir(DevicePath);
	if (DeviceDirectory != NULL)
	{
		char DeviceName[1024];
		struct dirent	*DeviceEntry;
		while (((DeviceEntry = readdir(DeviceDirectory)) != NULL) && (tGraphicsDevice == NULL))
		{
			if (strcmp(DeviceEntry->d_name, ".") && strcmp(DeviceEntry->d_name, "..") && strcmp(DeviceEntry->d_name, "stub"))
			{
//				cout << DeviceCount << ": " << DeviceEntry->d_name << endl;
				if (DeviceCount == DeviceIndex)
				{
					//found requested device, try to open it
					strcpy(DeviceName, DevicePath);
					strcat(DeviceName, "/");
					strcat(DeviceName, DeviceEntry->d_name);
					tGraphicsDevice = new GraphicsDevice(DeviceName);
				}
				DeviceCount ++;
			}
		}
		closedir(DeviceDirectory);
	}

	if (tGraphicsDevice != NULL)
	{
		if (tGraphicsDevice->InitCheck())
		{
			//this graphics card is supported
			return tGraphicsDevice;
		}
		else
		{
			//device not found or not supported 
			delete tGraphicsDevice;
			return NULL;
		}
	}
	else
		return NULL;
}

uint32 DeviceManager::GraphicsDeviceCount(uint32 *first_working_kerneldriver)
{
	uint32 DeviceCount(0);
	uint32 Index(0);
	const char * DevicePath = "/dev/graphics";
	DIR * DeviceDirectory;
	DeviceDirectory = opendir(DevicePath);

	//preset no working driver found
	*first_working_kerneldriver = 0xffffffff;

	if (DeviceDirectory != NULL)
	{
		char DeviceName[1024];
		struct dirent	*DeviceEntry;
		int hDevice (-1);
		bool noted_first_ok_drv(false);

		bool AccExists(false);
		char
			signature[1024],
			path[PATH_MAX];
		struct stat st;
		static const directory_which vols[] =	
		{
			B_BEOS_ADDONS_DIRECTORY,
			B_COMMON_ADDONS_DIRECTORY,
			B_USER_ADDONS_DIRECTORY
		};

		while ((DeviceEntry = readdir(DeviceDirectory)) != NULL)
		{
			if (strcmp(DeviceEntry->d_name, ".") && strcmp(DeviceEntry->d_name, "..")
					&& strcmp(DeviceEntry->d_name, "stub") && strcmp(DeviceEntry->d_name, "vesa"))
			{
				//open the driver: various drivers may have various ways to 'tell' us they won't work because
				//they weren't coldstarted...
				//(confirmed SiS (VendorID 0x1039) for 6326 AGP-8Mb (cardID 0x6326) using Be driver (R5/dano).
				strcpy(DeviceName, DevicePath);
				strcat(DeviceName, "/");
				strcat(DeviceName, DeviceEntry->d_name);
				hDevice = open(DeviceName, B_READ_WRITE);
				if (hDevice >= 0)
				{
					Log("DeviceManager::GfxDevCnt: kerneldriver '%s' opened OK\n", DeviceEntry->d_name);

					//--let's check if the driver 'is complete', but DONT use it(!)--
					//get the accelerant signature from the kerneldriver...
					if (ioctl(hDevice, B_GET_ACCELERANT_SIGNATURE, &signature, sizeof(signature)) == B_OK)
					{
						for (uint i = 0; i < sizeof (vols) / sizeof (vols[0]); i++)
						{
							//	compute directory path to common or beos addon directory on
							//	floppy or boot volume
  						if (find_directory (vols[i], -1, false, path, PATH_MAX) != B_OK) continue;

							strcat (path, "/accelerants/");
							strcat (path, signature);
							Log("DeviceManager::GfxDevCnt: Searching for acc. with signature: %s\n", signature);

							// is a file here with the correct signature (name)?
							if (stat(path, &st) == 0)
							{
								//yes..
								AccExists = true;
								break;
							}
						}
					}

					if (AccExists)
					{
						Log("DeviceManager::GfxDevCnt: Accelerant exists.\n");

						//report the first working driver's 'index' so the node won't be using it
						//(it displays the Desktop, hopefully)
						if (!noted_first_ok_drv)
						{
							*first_working_kerneldriver = Index;
							noted_first_ok_drv = true;
						}
						//device kerneldriver has been opened sucessfully, so count it
						DeviceCount ++;
						//and close it again
						close(hDevice);
						hDevice = -1;	
					}
					else
						Log("DeviceManager::GfxDevCnt: Accelerant does not exist.\n");
				}
				else
				{
					//we must assume the first kerneldriver being OK (for max 'broken Be drivers' compatibility):
					//'good designed' kerneldrivers should not allow a second open; cloning an accelerant
					//does not require that!! (see comments below.)
					if (!Index)
					{
						//report the first working driver's 'index' so the node won't be using it
						//(it displays the Desktop, hopefully)
						*first_working_kerneldriver = 0;
						noted_first_ok_drv = true;

						DeviceCount++;
					}

					//- kerneldriver is already in use by some first instance of it's accelerant:
					//  cloning is done without opening the kerneldriver another time (example:
					//  Be unified Matrox Mystique - G200 driver);
					//- (Be) kerneldriver is 'reporting' that the card we try to interface with is currently
					//  disfunctional:
					//  i.e. the card needs to be coldstarted, but the driver lacks this functionality; the
					//  card wasn't coldstarted by the system BIOS either (it's not the primary card!).
					//  (example: Be SiS 6326 driver).
					//
					// Note please:
					// - The first restriction (already in use) should best be 'restored' in all our gfx kernel
					//   drivers:
					//   it's a fast, easy and _very_ safe way to find out if a certain card is still free!
					// - The latter restriction should also be restored: but *in the accelerant* as the
					//   'not coldstarted' condition can best be checked there AND this gives optimum Be
					//   compatibility options for 'clients' like this node.
					//   (The driver-coldstart does not happen until the accelerant is inited!)
					//   (The accelerant's 'Init_accelerant' and 'clone_accelerant' fu's return a status_t...)
					//
					// This 'kerneldriver signalling' apparantly 'went missing' since Be 'introduced' the
					// alpha R4 graphics driver kit:
					// However, you'll find a source-code kerneldriver 'example' in the 'save' directory in the
					// same kit (which is part of the Be unified Matrox Mystique - G200 driver :).
					// Kit version: alpha release 2, 1999-03-30.
					Log("DeviceManager::GfxDevCnt: kerneldriver '%s' is blocked for open!\n",
						DeviceEntry->d_name);
				}
				Index++;
			}
		}
		closedir(DeviceDirectory);
	}
	return DeviceCount;
}
