//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2006.01.14
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson
//              2005/2006 Rudolf Cornelissen

//VGA Device Output consumer node, 18 january 2006, alpha 3 release.


#include "GraphicsDevice.h"
#include "Log.h"

#include <iostream.h>

GraphicsDevice::GraphicsDevice(const char * DeviceName)
:
	iGraphicsDeviceMutex(new BLocker),
	iDeviceName(DeviceName),
	iInitCheck(false),
	hDevice(-1),
	AccelerantImage(-1),
	GetAccelerantFunction(NULL),
	AccelerantDeviceInfo(new accelerant_device_info)
{
	Log("GraphicsDevice::GraphicsDevice()\n");

	if (iGraphicsDeviceMutex == NULL)
	{
		Log("GraphicsDevice::GraphicsDevice ERROR: BLocker iGraphicsDeviceMutex failed.\n");
		return;
	}
	
	BAutolock tAutolocker( iGraphicsDeviceMutex );

	//open the driver...
	hDevice = open(DeviceName, B_READ_WRITE);
	if (hDevice >= 0)
	{
		//device driver has been opened sucessfully...

		//load the accelerant image...
		AccelerantImage = LoadAccelerantImage();
		if (AccelerantImage >= 0)
		{
			//accelerant image has been loaded successfully...
			iInitCheck = true;
			iGetMode();
		}
		else
		{
			close(hDevice);
			hDevice = -1;	
		}
	}
}


GraphicsDevice::~GraphicsDevice()
{
	Log("GraphicsDevice::~GraphicsDevice()\n");

	if (iGraphicsDeviceMutex != NULL)
		iGraphicsDeviceMutex->Lock();

	delete AccelerantDeviceInfo;

	//unload the accelerant image...
	if (AccelerantImage >= 0)
	{
		//tell the driver to turn-off the monitor
		drv_set_dpms_mode(B_DPMS_STAND_BY);

		//un-initialize the accelerant...
		uninit_accelerant UnInitAccelerant(NULL);
		UnInitAccelerant = (uninit_accelerant)GetAccelerantFunction(B_UNINIT_ACCELERANT, NULL);
		if (UnInitAccelerant != NULL)
			UnInitAccelerant();
		
		//unload the accelerant image...
		unload_add_on(AccelerantImage);
	}
	
	//close the driver...
	if (hDevice >= 0)
		close(hDevice);
		
	if (iGraphicsDeviceMutex != NULL)
	{
//		iGraphicsDeviceMutex->Unlock();
		delete iGraphicsDeviceMutex;
		iGraphicsDeviceMutex = NULL;
	}
}


bool GraphicsDevice::InitCheck()
{
	return iInitCheck;
}


const char * GraphicsDevice::DeviceName()
{
	return iDeviceName;
}


bool GraphicsDevice::Lock()
{
	return iGraphicsDeviceMutex->Lock();
}


status_t GraphicsDevice::LockWithTimeout( bigtime_t uTimeout )
{
	return iGraphicsDeviceMutex->LockWithTimeout( uTimeout );
}


void GraphicsDevice::Unlock()
{
	iGraphicsDeviceMutex->Unlock();
}


const char * GraphicsDevice::Name()
{
	return AccelerantDeviceInfo->name;
}


const char * GraphicsDevice::Chipset()
{
	return AccelerantDeviceInfo->chipset;
}


const char * GraphicsDevice::SerialNumber()
{
	return AccelerantDeviceInfo->serial_no;
}


uint32 GraphicsDevice::Memory()
{
	return AccelerantDeviceInfo->memory;
}


uint32 GraphicsDevice::DACSpeed()
{
	return AccelerantDeviceInfo->dac_speed;
}


image_id GraphicsDevice::LoadAccelerantImage()
{
	status_t result;
	image_id image = -1;
	uint i;
	char
		signature[1024],
		path[PATH_MAX];
	struct stat st;
	static const directory_which vols[] =	
	{
		B_BEOS_ADDONS_DIRECTORY,
		B_COMMON_ADDONS_DIRECTORY,
		B_USER_ADDONS_DIRECTORY
	};

	//get the accelerant signature from the driver...
	result = ioctl(hDevice, B_GET_ACCELERANT_SIGNATURE, &signature, sizeof(signature));
	if (result != B_OK) goto done;

	for (i=0; i < sizeof (vols) / sizeof (vols[0]); i++)
	{
		//	compute directory path to common or beos addon directory on
		//	floppy or boot volume

  	if (find_directory (vols[i], -1, false, path, PATH_MAX) != B_OK) continue;

		strcat (path, "/accelerants/");
		strcat (path, signature);
		Log("Searching for accelerant with signature: %s\n", signature);

		// don't try to load non-existant files
		if (stat(path, &st) != 0) continue;

		// load the image
		Log("trying to load accelerant image\n");
		image = load_add_on(path);
		if (image >= 0)
		{
			// get entrypoint from accelerant
			result = get_image_symbol (image,B_ACCELERANT_ENTRY_POINT,	B_SYMBOL_TYPE_ANY, (void **)&GetAccelerantFunction);
			if (result == B_OK)
			{
				drv_init_accelerant = (init_accelerant)GetAccelerantFunction(B_INIT_ACCELERANT, NULL);
				if (drv_init_accelerant && ((result = drv_init_accelerant(hDevice)) == B_OK))
				{
					if (GetDriverHooks() == B_OK)
					{
						// we have a winner!
						break;
					}
					else
					{
						//un-initialize the accelerant...
						uninit_accelerant UnInitAccelerant(NULL);
						UnInitAccelerant = (uninit_accelerant)GetAccelerantFunction(B_UNINIT_ACCELERANT, NULL);
						if (UnInitAccelerant != NULL)
							UnInitAccelerant();
					}
				}
			}
			// unload the accelerant, as we must be able to init!
			Log("unloading accelerant image\n");
			unload_add_on(image); 
		}
		// mark failure to load image
		image = -1;
	}

done:
	if (image >= 0)
			Log("Accelerant is useable\n");
	else
			Log("Accelerant is not useable\n");

	return image;
}

status_t GraphicsDevice::GetDriverHooks()
{
	bool virt_screen_supported = false;

	drv_vid_caps = 0;

	//required material	
	drv_propose_display_mode = (propose_display_mode)GetAccelerantFunction(B_PROPOSE_DISPLAY_MODE, NULL);
	drv_set_display_mode = (set_display_mode)GetAccelerantFunction(B_SET_DISPLAY_MODE, NULL);
	drv_get_display_mode = (get_display_mode)GetAccelerantFunction(B_GET_DISPLAY_MODE, NULL);
	drv_get_frame_buffer_config = (get_frame_buffer_config)GetAccelerantFunction(B_GET_FRAME_BUFFER_CONFIG, NULL);
	drv_dpms_capabilities = (dpms_capabilities)GetAccelerantFunction(B_DPMS_CAPABILITIES, NULL);
	drv_set_dpms_mode = (set_dpms_mode)GetAccelerantFunction(B_SET_DPMS_MODE, NULL);
	drv_get_accelerant_device_info = (get_accelerant_device_info)GetAccelerantFunction(B_GET_ACCELERANT_DEVICE_INFO, NULL);

	if (!drv_propose_display_mode || !drv_set_display_mode || !drv_get_display_mode ||
		  !drv_get_frame_buffer_config || !drv_dpms_capabilities || !drv_set_dpms_mode || !drv_get_accelerant_device_info)
	  return B_ERROR;

	//we want to keep the monitor shut-off, so we need to make sure the driver supports that
	drv_dpms_caps = drv_dpms_capabilities();
	if (!(drv_dpms_caps & B_DPMS_STAND_BY))
		return B_ERROR;

	//now tell the driver to keep the monitor shut-off (otherwise the first SetMode call will turn it on)
	drv_set_dpms_mode(B_DPMS_STAND_BY);

	//set a display_mode now, because the other functions may or may not be supported on a per-mode basis
	//setting 640x480x32 @ 70Hz (standard VESA mode)
	iDisplayMode.virtual_width = 640;
	//let's see if this virtual_height mode is still accelerated so we can do at least singlebuffered scaled_blits...
	iDisplayMode.virtual_height = 480 * 2;
	iDisplayMode.h_display_start = 0;
	iDisplayMode.v_display_start = 0;
	iDisplayMode.flags = 0;
	iDisplayMode.space = B_RGB32;
	iDisplayMode.timing.h_display = 640;
	iDisplayMode.timing.h_sync_start = 664;
	iDisplayMode.timing.h_sync_end = 704;
	iDisplayMode.timing.h_total = 832;
	iDisplayMode.timing.v_display = 480;
	iDisplayMode.timing.v_sync_start = 489;
	iDisplayMode.timing.v_sync_end = 492;
	iDisplayMode.timing.v_total = 520;
	iDisplayMode.timing.pixel_clock = ((832 * 520 * 70) / 1000);
	iDisplayMode.timing.flags = 0;
	//see if this device can do the mode as requested (including virtual size)..
	//(must make a copy as iDisplayMode will be modified by Proposemode!!!)
	display_mode fBounds = iDisplayMode;
	if (ProposeMode(&iDisplayMode, &fBounds, &fBounds) == B_OK)
	{
		//can do.. so we also have enough memory to use virtual_screen scaled_blit method for drawing!
		virt_screen_supported = true;
	}
	else
	{
		//can't do, set a normal mode (non-virtual)
		iDisplayMode.virtual_height = 480;
		//let's save some memory (overlay doesn't care about the desktop space)
		iDisplayMode.space = B_CMAP8;
	}
	//various drivers may have various ways to 'tell' us they won't work because they weren't coldstarted...
	if (SetMode(&iDisplayMode) != B_OK) return B_ERROR;

	//now tell the driver again to keep the monitor shut-off:
	//although Be's R4 alpha graphics driver kit tells us that a setmode command should retain our
	//current DPMS mode, most Be drivers don't adhere to that but force B_DPMS_ON state, turning on the screen.
	//Please note:
	//The R4 kit's way seems the way to go here. (Apparantly Be did not find the time to update
	//all of it's drivers :-)
	//Tested Be drivers: Nvidia and Matrox drivers: 'faulty', SiS 6326 driver: OK.
	drv_set_dpms_mode(B_DPMS_STAND_BY);

	drv_vid_caps = 0;
	//required for acceleration mode
	drv_acquire_engine = (acquire_engine)GetAccelerantFunction(B_ACQUIRE_ENGINE, NULL);
	drv_release_engine = (release_engine)GetAccelerantFunction(B_RELEASE_ENGINE, NULL);
	drv_wait_engine_idle = (wait_engine_idle)GetAccelerantFunction(B_WAIT_ENGINE_IDLE, NULL);
	//this function won't get exported if the mode is unaccelerated...
	drv_screen_to_screen_scaled_filtered_blit =
		(screen_to_screen_scaled_filtered_blit)GetAccelerantFunction(B_SCREEN_TO_SCREEN_SCALED_FILTERED_BLIT, NULL);
	if (drv_acquire_engine && drv_release_engine && drv_wait_engine_idle &&
			drv_screen_to_screen_scaled_filtered_blit && virt_screen_supported)
	  drv_vid_caps = VID_BLIT;

	//required for overlay mode
	drv_overlay_supported_spaces = (overlay_supported_spaces)GetAccelerantFunction(B_OVERLAY_SUPPORTED_SPACES, NULL);
	drv_allocate_overlay_buffer = (allocate_overlay_buffer)GetAccelerantFunction(B_ALLOCATE_OVERLAY_BUFFER, NULL);
	drv_release_overlay_buffer = (release_overlay_buffer)GetAccelerantFunction(B_RELEASE_OVERLAY_BUFFER, NULL);
	drv_allocate_overlay = (allocate_overlay)GetAccelerantFunction(B_ALLOCATE_OVERLAY, NULL);
	drv_release_overlay = (release_overlay)GetAccelerantFunction(B_RELEASE_OVERLAY, NULL);
	drv_configure_overlay = (configure_overlay)GetAccelerantFunction(B_CONFIGURE_OVERLAY, NULL);
	if (drv_overlay_supported_spaces && drv_allocate_overlay_buffer && drv_release_overlay_buffer &&
			drv_allocate_overlay && drv_release_overlay && drv_configure_overlay)
	  drv_vid_caps |= VID_OVERLAY;


	//we need at least one method to draw
	if (!drv_vid_caps) return B_ERROR;

	//we 'need' to fetch this once _somewhere_..
	drv_get_accelerant_device_info(AccelerantDeviceInfo);
	//...and do it again: the Be unified Nvidia driver (at least) is faulty here: it gives back
	//   correct info after the _second_ time (or later) that we invoke it!
	drv_get_accelerant_device_info(AccelerantDeviceInfo);

	//let's log some stuff...
	if (drv_vid_caps & VID_OVERLAY)
		Log("Accelerant supports OVERLAY drawing mode\n");
	if (drv_vid_caps & VID_BLIT)
		Log("Accelerant supports WORKSPACE SCALED BLIT drawing mode\n");
	Log("Card memory size reported: %.1fMb\n", (AccelerantDeviceInfo->memory / (1024. * 1024.)));

	//optional
	retrace_sem = -1;
	drv_accelerant_retrace_semaphore =
		(accelerant_retrace_semaphore)GetAccelerantFunction(B_ACCELERANT_RETRACE_SEMAPHORE, NULL);
	if (drv_accelerant_retrace_semaphore)
		retrace_sem = (sem_id)drv_accelerant_retrace_semaphore();

	return B_OK;
}

uint32 GraphicsDevice::GetDrawModes()
{
	return drv_vid_caps;
}

status_t GraphicsDevice::GetMode( display_mode * uDisplayMode )
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	status_t result(B_ERROR);
	result = iGetMode();
	if (result >= B_OK)
		*uDisplayMode = iDisplayMode;
	return result;
}


status_t GraphicsDevice::iGetMode()
{
	status_t result(B_ERROR);
	result = drv_get_display_mode(&iDisplayMode);
	if (result >= B_OK)
	{
		if (drv_get_frame_buffer_config) drv_get_frame_buffer_config(&iFrameBufferConfig);
	}
	return result;
}


status_t GraphicsDevice::SetMode( display_mode * new_mode, bool)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	
	if (drv_set_display_mode)
	{
		status_t result(B_ERROR);
		result = drv_set_display_mode( new_mode );
		if (result >= B_OK)
		{
			iDisplayMode = *new_mode;
			if (drv_get_frame_buffer_config) drv_get_frame_buffer_config(&iFrameBufferConfig);
			Log("New Mode Set OK!\n");
		}
		return result;
	}
	return B_ERROR;
}


status_t GraphicsDevice::ProposeMode(
	display_mode *target, const display_mode *low, const display_mode *high )
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	if (drv_propose_display_mode)
		return drv_propose_display_mode( target, const_cast<display_mode*>(low), const_cast<display_mode*>(high) );
	else
		return B_ERROR;
}


status_t GraphicsDevice::GetFrameBufferConfig(
	frame_buffer_config &uFrameBufferConfig ) const
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	uFrameBufferConfig = iFrameBufferConfig;
	return B_OK;
}

status_t GraphicsDevice::AcquireEngine()
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	eng_token = NULL;
	return drv_acquire_engine(B_2D_ACCELERATION, 0, NULL, &eng_token);
}

status_t GraphicsDevice::ReleaseEngine()
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	if (eng_token)
		return drv_release_engine(eng_token, NULL);
	else
		return B_ERROR;
}

void GraphicsDevice::WaitEngineIdle()
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	drv_wait_engine_idle();
}

void GraphicsDevice::ScaledBlit(const int32 uBufferFlags)
{
//	cout << "bufferflags is: " << uBufferFlags << endl;

	BAutolock tAutolocker( iGraphicsDeviceMutex );
	drv_screen_to_screen_scaled_filtered_blit(eng_token, &(blit_list[uBufferFlags]), 1);
}

status_t GraphicsDevice::WaitForRetrace()
{
	if (retrace_sem >= 0)
	{
		acquire_sem(retrace_sem);
		//note: the graphics kerneldriver releases the sem, so we can acquire it again.
		return B_OK;
	}
	return B_ERROR;
}

status_t GraphicsDevice::GetOverlayBufferConfig(const overlay_buffer *mob, uint16 *bytes_per_row ) const
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	*bytes_per_row = mob->bytes_per_row;
	return B_OK;
}

void GraphicsDevice::AllocateOverlay()
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	my_overlay_token = drv_allocate_overlay();
}


//SiS 6326 driver notes:
//this function is broken in Be's SiS accelerant (tested 6326 AGP 8Mb): overlay is NOT shutdown...
//(that same driver configures overlay faulty as well, you'll find that it displays vertical bars of
// more or less repeating video, instead of normal video. Oh, and it only supports single buffered
// overlay: but that's no fault: it's a valid 'option'.
// And another thing: this driver reports having 0.0Mb graphics RAM in 'get_accelerant_device_info'...)
status_t GraphicsDevice::ReleaseOverlay()
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	return drv_release_overlay(my_overlay_token);
}

const overlay_buffer* GraphicsDevice::AllocateOverlayBuffer(uint16 width, uint16 height)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	return drv_allocate_overlay_buffer(B_YCbCr422, width, height);
}

status_t GraphicsDevice::ReleaseOverlayBuffer(const overlay_buffer *ob)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	return drv_release_overlay_buffer(ob);
}

status_t GraphicsDevice::ConfigureOverlay(const int32 uBufferFlags)
{
//	cout << "bufferflags is: " << uBufferFlags << endl;

	BAutolock tAutolocker( iGraphicsDeviceMutex );
	return drv_configure_overlay(my_overlay_token, my_overlay_buffer[uBufferFlags], &my_overlay_window, &my_overlay_view);
}

void GraphicsDevice::SetDPMS(const uint32 flags)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	drv_set_dpms_mode(flags);
}

void GraphicsDevice::ClearScreen()
{
	//using 64-bit writes for the fastest write-actions possible,
	//assuming screen memsize has size granularity of 8 bytes or higher...
	uint64 *adress = (uint64 *)(iFrameBufferConfig.frame_buffer);
	uint64 data = 0x0000000000000000;
	for (uint32 offset = 0; offset < ((iFrameBufferConfig.bytes_per_row * 480) >> 3); offset ++)
	{
		adress[offset] = data;
	}
}

BBufferGroup * GraphicsDevice::CreateOverlayBufferGroup(uint8 *uDesiredBufferCount, const media_format &format)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	if (*uDesiredBufferCount > MaxBuffers) *uDesiredBufferCount = MaxBuffers;
	area_id MyFrameBuffer;
	
	MyFrameBuffer = area_for(iFrameBufferConfig.frame_buffer);
	if (MyFrameBuffer >= 0)
	{
		Log("Framebuffer Area found.\n");
	}
	else
		return NULL;

	area_info ai;
	if (MyFrameBuffer >= 0)
	{
		//Log("Framebuffer Area info:\n");
		get_area_info(MyFrameBuffer, &ai);
		//cout << "  Area ID: " << ai.area << endl;
		//cout << "  Name: " << ai.name << endl;
		//cout << "  Size (K): " << (ai.size / 1024) << endl;
		//cout << "  Locked Size (K): " << (ai.ram_size / 1024) << endl;
		//cout << "  Team Address: " << ai.address << endl;
		//cout << "  FB Address: " << iFrameBufferConfig.frame_buffer << endl;
		//cout << "  FB Offset: " << (((char *)iFrameBufferConfig.frame_buffer) - ((char *)ai.address)) << endl;
	}
	else
	{
		Log(strerror(MyFrameBuffer));
		Log("ERROR: Framebuffer Area could not be found or created.\n");
		return NULL;
	}
	
	int BufferCount(0);
	size_t BufferOffset;
	BBufferGroup *my_group = new BBufferGroup;

	//update mode if needed to spare RAM
	if (iDisplayMode.virtual_height != 480)
	{
		//no virtual screen needed as we are using overlay
		iDisplayMode.virtual_height = 480;
		//with overlay the depth of the workspace is unimportant
		iDisplayMode.space = B_CMAP8;
		SetMode(&iDisplayMode);
	}

	if (drv_vid_caps & VID_OVERLAY)
	{
		while (BufferCount < *uDesiredBufferCount)
		{
			my_overlay_buffer [BufferCount] = AllocateOverlayBuffer(
				format.u.raw_video.display.line_width, format.u.raw_video.display.line_count);

			if (my_overlay_buffer [BufferCount])
			{
				BufferOffset = ((char *)(my_overlay_buffer [BufferCount]->buffer)) - ((char *)ai.address);

				buffer_clone_info bc_info;
				bc_info.area = MyFrameBuffer;
				bc_info.offset = BufferOffset;
				bc_info.size = format.u.raw_video.display.bytes_per_row * format.u.raw_video.display.line_count;

				//works ok, although not really officially meant to be used like this..
				//(officially this is a fieldcounter and also holds BSmallBuffer flag)
				bc_info.flags = BufferCount;
			
				if (my_group->AddBuffer(bc_info) == B_OK)
				{
					BufferCount++;
					Log("New buffer added OK\n");
				}
				else
				{
					Log("Could not add new buffer!!!\n");
					break;
				}
			}
			else
			{
				Log("Could not add new buffer (driver limitation)\n");
				break;
			}
		}
		//set requested number of buffers to actually created buffers
		*uDesiredBufferCount = BufferCount;
		if (!*uDesiredBufferCount)
		{
			Log("Could not create a single new buffer, aborting\n");
			delete my_group;
			return NULL;
		}

		//setup overlay output window properties
		my_overlay_window.h_start = 0;
		my_overlay_window.v_start = 0;
		my_overlay_window.width = 640;
		my_overlay_window.height = 480;

		my_overlay_window.offset_left = 0;
		my_overlay_window.offset_top = 0;
		my_overlay_window.offset_right = 0;
		my_overlay_window.offset_bottom = 0;

		my_overlay_window.flags = (B_OVERLAY_HORIZONTAL_FILTERING | B_OVERLAY_VERTICAL_FILTERING);
	}

	return my_group;
}

void GraphicsDevice::SetScaledBlitList(const int uDesiredBufferCount, const media_format &format)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );

	int tDesiredBufferCount = uDesiredBufferCount;
	if (tDesiredBufferCount > MaxBuffers) tDesiredBufferCount = MaxBuffers;

	int BufferCount(0);
	while (BufferCount < tDesiredBufferCount)
	{
			//confirmed OK:
			blit_list[BufferCount].src_left = 0;
			blit_list[BufferCount].src_top = (BufferCount + 1) * 480;
			blit_list[BufferCount].src_width = format.u.raw_video.display.line_width - 1;
			blit_list[BufferCount].src_height = format.u.raw_video.display.line_count - 1;
			blit_list[BufferCount].dest_left = 0;
			blit_list[BufferCount].dest_top = 0;
			blit_list[BufferCount].dest_width = 639;
			blit_list[BufferCount].dest_height = 479;

			BufferCount++;
	}
}

//note: Zoom can be used to display only a part of a overlay buffer still filling up the output window...
void GraphicsDevice::SetOverlayZoom(const int /*uDesiredBufferCount*/, const media_format &format)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );

	my_overlay_view.h_start = 0;
	my_overlay_view.v_start = 0;
	my_overlay_view.width = format.u.raw_video.display.line_width;
	my_overlay_view.height = format.u.raw_video.display.line_count;
}

BBufferGroup * GraphicsDevice::CreateFixedBlitBufferGroup(uint8 *uDesiredBufferCount)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	if (*uDesiredBufferCount > MaxBuffers) *uDesiredBufferCount = MaxBuffers;
	area_id MyFrameBuffer;

	MyFrameBuffer = area_for(iFrameBufferConfig.frame_buffer);
	if (MyFrameBuffer >= 0)
	{
		Log("Framebuffer Area found.\n");
	}
	else
		return NULL;

	area_info ai;
	if (MyFrameBuffer >= 0)
	{
		//Log("Framebuffer Area info:\n");
		get_area_info(MyFrameBuffer, &ai);
		//cout << "  Area ID: " << ai.area << endl;
		//cout << "  Name: " << ai.name << endl;
		//cout << "  Size (K): " << (ai.size / 1024) << endl;
		//cout << "  Locked Size (K): " << (ai.ram_size / 1024) << endl;
		//cout << "  Team Address: " << ai.address << endl;
		//cout << "  FB Address: " << iFrameBufferConfig.frame_buffer << endl;
		//cout << "  FB Offset: " << (((char *)iFrameBufferConfig.frame_buffer) - ((char *)ai.address)) << endl;

	}
	else
	{
		Log(strerror(MyFrameBuffer));
		Log("ERROR: Framebuffer Area could not be found or created.\n");
		return NULL;
	}

	display_mode fBounds;
	int BufferCount(0);
	size_t BufferOffset;
	BBufferGroup *my_group = new BBufferGroup;

	BufferCount = 0;
	if (drv_vid_caps & VID_BLIT)
	{
		while (BufferCount < *uDesiredBufferCount)
		{
			//with blits the depth of the workspace should be optimal
			iDisplayMode.space = B_RGB32;
			//update requested mode to have room for our buffer(s)
			iDisplayMode.virtual_height = (480 * (BufferCount + 2));
			//check if this virtual mode can be set AND if we still have acceleration..
			//(must make a copy as iDisplayMode will be modified by Proposemode!!!)
			fBounds = iDisplayMode;
			if ((ProposeMode(&iDisplayMode, &fBounds, &fBounds) == B_OK) &&
					(SetMode(&iDisplayMode) == B_OK) &&
					GetAccelerantFunction(B_SCREEN_TO_SCREEN_SCALED_FILTERED_BLIT, NULL))
			{
				BufferOffset = ((char *)iFrameBufferConfig.frame_buffer) - ((char *)ai.address);
				BufferOffset += (iFrameBufferConfig.bytes_per_row * 480 * (BufferCount + 1));

				buffer_clone_info bc_info;
				bc_info.area = MyFrameBuffer;
				bc_info.offset = BufferOffset;
				bc_info.size = iFrameBufferConfig.bytes_per_row * 480;

				//works ok, although not really officially meant to be used like this..
				//(officially this is a fieldcounter and also holds BSmallBuffer flag)
				bc_info.flags = BufferCount;
			
				if (my_group->AddBuffer(bc_info) == B_OK)
				{
					BufferCount++;
					Log("New buffer added OK\n");
				}
				else
				{
					Log("Could not add new buffer!!!\n");
					break;
				}
			}
			else
			{
				Log("Could not add new buffer (driver limitation)\n");
				break;
			}
		}

		//update requested mode to fit with created number of buffer(s)
		iDisplayMode.virtual_height = (480 * (BufferCount + 1));
		//set it
		SetMode(&iDisplayMode);

		//set requested number of buffers to actually created buffers
		*uDesiredBufferCount = BufferCount;
		if (!*uDesiredBufferCount)
		{
			Log("Could not create a single new buffer, aborting\n");
			delete my_group;
			return NULL;
		}
	}

	return my_group;
}

void GraphicsDevice::DeleteBlitBufferGroup(BBufferGroup * uBufferGroup)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );
	if (uBufferGroup != NULL)
	{
		delete uBufferGroup;
		//would be nice if this would work, but it doesn't!!! (do it in calling code instead)
		uBufferGroup = NULL;
	}
}

void GraphicsDevice::DeleteOverlayBufferGroup(BBufferGroup * uBufferGroup)
{
	BAutolock tAutolocker( iGraphicsDeviceMutex );

	int32 bufs;
	if (uBufferGroup != NULL)
	{
		if (uBufferGroup->CountBuffers(&bufs) == B_OK)
		{
			for (int32 cnt = 0; cnt < bufs; cnt++)
				if (drv_vid_caps & VID_OVERLAY) ReleaseOverlayBuffer(my_overlay_buffer [cnt]);
		}

		delete uBufferGroup;
		//would be nice if this would work, but it doesn't!!! (do it in calling code instead)
		uBufferGroup = NULL;
	}
}
