//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2006.01.14
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002      Kevin H. Patterson
//              2005-2006 Rudolf Cornelissen

#ifndef _GraphicsDevice_h
#define _GraphicsDevice_h

	#include <Application.h>
	#include <Locker.h>
	#include <Autolock.h>
	#include "video_overlay.h"

	//For 'slimdemo' code
		#include <OS.h>
		#include <image.h>
		#include <FindDirectory.h>
		#include <graphic_driver.h>
		#include <Accelerant.h>
		#include <dirent.h>
		#include <stdio.h>
		#include <string.h>
		#include <sys/stat.h>
		#include <malloc.h>
		#include <errno.h>
		#include <stdlib.h>

	//For 'MediaNode' code
		//#include <Area.h>
		#include <Buffer.h>
		#include <BufferGroup.h>

		//for drv_vid_caps
		enum {
			VID_BLIT = 1 << 0,
			VID_OVERLAY = 1 << 1
		};

	class GraphicsDevice
	{
	friend class DeviceManager;
	private:
		GraphicsDevice(const char * DeviceName);
		bool InitCheck();

	public:
		bool Lock();
		status_t LockWithTimeout( bigtime_t uTimeout );
		void Unlock();

		~GraphicsDevice();
		const char * DeviceName();
		const char * Name();
		const char * Chipset();
		const char * SerialNumber();
		uint32 Memory();
		uint32 DACSpeed();

		status_t GetDriverHooks();
		uint32 GetDrawModes();

		status_t GetMode( display_mode * uDisplayMode );
		status_t SetMode( display_mode * uDisplayMode, bool uMakeDefault = false );
		status_t ProposeMode(
			display_mode *target,
			const display_mode *low, const display_mode *high );
			
		status_t GetFrameBufferConfig( frame_buffer_config &uFrameBufferConfig ) const;

		status_t GraphicsDevice::AcquireEngine();
		status_t GraphicsDevice::ReleaseEngine();
		void GraphicsDevice::WaitEngineIdle();
		void GraphicsDevice::ScaledBlit(const int32 uBufferFlags);
		status_t GraphicsDevice::ConfigureOverlay(const int32 uBufferFlags);
		void GraphicsDevice::SetDPMS(const uint32 flags);

		status_t WaitForRetrace();
		
		status_t GraphicsDevice::GetOverlayBufferConfig(const overlay_buffer *mob, uint16 *bytes_per_row ) const;
		void GraphicsDevice::AllocateOverlay();
		status_t GraphicsDevice::ReleaseOverlay();
		const overlay_buffer* GraphicsDevice::AllocateOverlayBuffer(uint16 width, uint16 height);
		status_t GraphicsDevice::ReleaseOverlayBuffer(const overlay_buffer *ob);


		void GraphicsDevice::ClearScreen();
		BBufferGroup * CreateOverlayBufferGroup(uint8 *uDesiredBufferCount, const media_format &format);
		BBufferGroup * CreateFixedBlitBufferGroup(uint8 *uDesiredBufferCount);
		void SetScaledBlitList(const int uDesiredBufferCount,	const media_format &format);
		void SetOverlayZoom(const int uDesiredBufferCount,	const media_format &format);
		void DeleteBlitBufferGroup(BBufferGroup * uBufferGroup);
		void DeleteOverlayBufferGroup(BBufferGroup * uBufferGroup);

	private:
		status_t iGetMode();
		
	private:
		image_id LoadAccelerantImage();
		
	private:
		BLocker * iGraphicsDeviceMutex;
	
		const char * iDeviceName;
		bool iInitCheck;
		int hDevice;
		image_id AccelerantImage;
		GetAccelerantHook GetAccelerantFunction;
		accelerant_device_info * AccelerantDeviceInfo;
		
		display_mode iDisplayMode;
		frame_buffer_config iFrameBufferConfig;
		
		static const int MaxBuffers = 32;
		
		//acceleration:
		engine_token *eng_token;
		scaled_blit_params blit_list[3];

//driver hooks:
		init_accelerant drv_init_accelerant;
		get_frame_buffer_config drv_get_frame_buffer_config;
		propose_display_mode drv_propose_display_mode;
		set_display_mode drv_set_display_mode;
		get_display_mode drv_get_display_mode;
		dpms_capabilities drv_dpms_capabilities;
		set_dpms_mode drv_set_dpms_mode;
		get_accelerant_device_info drv_get_accelerant_device_info;

		acquire_engine drv_acquire_engine;
		release_engine drv_release_engine;
		wait_engine_idle drv_wait_engine_idle;
		screen_to_screen_scaled_filtered_blit drv_screen_to_screen_scaled_filtered_blit;

		overlay_supported_spaces drv_overlay_supported_spaces;
		allocate_overlay_buffer drv_allocate_overlay_buffer;
		release_overlay_buffer drv_release_overlay_buffer;
		allocate_overlay drv_allocate_overlay;
		release_overlay drv_release_overlay;
		configure_overlay drv_configure_overlay;

		accelerant_retrace_semaphore drv_accelerant_retrace_semaphore;

		uint32 drv_dpms_caps;
		sem_id retrace_sem;
		uint32 drv_vid_caps;
		const overlay_buffer* my_overlay_buffer[3];
		overlay_token my_overlay_token;
		overlay_window my_overlay_window;
		overlay_view my_overlay_view;
};
	
#endif
