// VideoAddOn.cpp
//VGA Device Output consumer node, 20 april 2006, alpha 3 release.

#include <OS.h>
#include <File.h>
#include "VideoAddOn.h"

#include "Log.h"

const char * VideoAddOn::kSettingsFileName =
    "/boot/home/config/settings/Media/VGADeviceOutput";

volatile port_id VideoAddOn::fVideoConsumerPort = 0;


VideoAddOn::VideoAddOn(
    image_id image)
 :  BMediaAddOn(image),
    fStatus(B_NO_INIT),
    fFlavors_count(0),
    fFormats_count(0),
    fConfigurations(),
    iDrvInfo(),
    fConsumer(NULL)
{
	static char name[32] = { "VGA Device Output (Alpha 3)" };
	static char info[80] = 
		{"VGA Device Output, (c)2002-2006 Kevin H. Patterson - Rudolf Cornelissen" };

	Log("VideoAddOn::VideoAddOn\n");

	/* Video Output parameters */
	fFlavors[0].name = name;
	fFlavors[0].info = info;
	fFlavors[0].kinds = B_BUFFER_CONSUMER | B_CONTROLLABLE | B_PHYSICAL_OUTPUT;
	fFlavors[0].flavor_flags = B_FLAVOR_IS_GLOBAL;
	fFlavors[0].internal_id = 7778;
	fFlavors[0].possible_count = 1;
	fFlavors[0].in_format_flags = 0;
	fFlavors[0].in_formats = fFormats;
	fFlavors[0].out_format_count = 0;
	fFlavors[0].out_format_flags = 0;
	fFlavors[0].out_formats = NULL;
	
	fFlavors_count = 1;

	//BETA: Some sort of check or other code should go here to see if the
	//video output hardware is available. fFlavors[].possible_count should
	//reflect the number of supported devices found.

	uint8 set_count = 0;
	DeviceManager *iDeviceManager = new DeviceManager;

	//preset no drivers found
	iDrvInfo.ok_nr_of_kerneldrvs = 0;
	for (uint8 cnt = 0; cnt < 32; cnt++)
		iDrvInfo.ok_and_free_acc_at[cnt].present = false;
	//iterate through all available graphics devices and get their names
	if (iDeviceManager != NULL)
	{
		GraphicsDevice * mGraphicsDevice(NULL);
		iDrvInfo.ok_nr_of_kerneldrvs =
			iDeviceManager->GraphicsDeviceCount(&(iDrvInfo.first_ok_kerneldrv));
		//do not use the first working kerneldriver:
		//it's (hopefully) in use by the system(!)
		if ((iDrvInfo.first_ok_kerneldrv != 0xffffffff) &&
			(iDrvInfo.ok_nr_of_kerneldrvs  >= 2))
		{
			for (uint32 idx(iDrvInfo.first_ok_kerneldrv + 1); idx < 32; idx++)
			{
				//GetGraphicsDevice also checks for a working kerneldriver...
				mGraphicsDevice = iDeviceManager->GetGraphicsDevice(idx);
				if (mGraphicsDevice != NULL)
				{
					iDrvInfo.ok_and_free_acc_at[idx].present = true;
					//src and dest are both 32 characters long:
					sprintf(iDrvInfo.ok_and_free_acc_at[idx].name, mGraphicsDevice->Name());
					//we need the drawing capabilities to determine valid colorspaces..
					iDrvInfo.ok_and_free_acc_at[idx].drw_modes = mGraphicsDevice->GetDrawModes();
				}
				delete mGraphicsDevice;
				mGraphicsDevice = NULL;
			}
		}
		delete iDeviceManager;
	}

	//now count useable and free _sets_ of kerneldrivers and accelerants:
	//we need at least _one_!!
	Log("VideoAddOn::VideoAddOn: Listing free drivers:\n");
	for (uint32 idx(0); idx < 32; idx++)
	{
		if (iDrvInfo.ok_and_free_acc_at[idx].present)
		{
			Log ("Useable drv_set #%d at index #%d with acc. name: %s\n",
				set_count, idx, iDrvInfo.ok_and_free_acc_at[idx].name);
			set_count++;
		}
	}
	Log("VideoAddOn::VideoAddOn: End of free drivers list.\n");

	//do we have a reason to keep running?
	if (set_count)
	{
		LoadConfigurationFile();
		fStatus = B_OK;
	}
	else
	{
		Log("VideoAddOn::VideoAddOn: Shutting down: no hardware available!\n");
		return;
	}

	/* Acceptable Video Formats */
	switch (GetSelectedDevCaps(&fConfigurations, &iDrvInfo))
	{
	case (VID_OVERLAY | VID_BLIT):
		Log("VideoAddOn::VideoAddOn: supporting B_YCbCr422, B_RGB32 and B_RGBA32\n");
		fFormats[0].type = B_MEDIA_RAW_VIDEO;
		fFormats[0].u.raw_video = media_raw_video_format::wildcard;
		fFormats[0].u.raw_video.interlace = 1;
		fFormats[0].u.raw_video.display.format = B_YCbCr422;

		fFormats[1].type = B_MEDIA_RAW_VIDEO;
		fFormats[1].u.raw_video = media_raw_video_format::wildcard;
		fFormats[1].u.raw_video.interlace = 1;
		fFormats[1].u.raw_video.display.format = B_RGB32;

		fFormats[2].type = B_MEDIA_RAW_VIDEO;
		fFormats[2].u.raw_video = media_raw_video_format::wildcard;
		fFormats[2].u.raw_video.interlace = 1;
		fFormats[2].u.raw_video.display.format = B_RGBA32;

		fFormats_count = 3;
		fFlavors[0].in_format_count = 3;
		break;
	case VID_OVERLAY:
		Log("VideoAddOn::VideoAddOn: supporting B_YCbCr422\n");
		fFormats[0].type = B_MEDIA_RAW_VIDEO;
		fFormats[0].u.raw_video = media_raw_video_format::wildcard;
		fFormats[0].u.raw_video.interlace = 1;
		fFormats[0].u.raw_video.display.format = B_YCbCr422;

		fFormats_count = 1;
		fFlavors[0].in_format_count = 1;
		break;
	case VID_BLIT:
		Log("VideoAddOn::VideoAddOn: supporting B_RGB32 and B_RGBA32\n");
		fFormats[0].type = B_MEDIA_RAW_VIDEO;
		fFormats[0].u.raw_video = media_raw_video_format::wildcard;
		fFormats[0].u.raw_video.interlace = 1;
		fFormats[0].u.raw_video.display.format = B_RGB32;

		fFormats[1].type = B_MEDIA_RAW_VIDEO;
		fFormats[1].u.raw_video = media_raw_video_format::wildcard;
		fFormats[1].u.raw_video.interlace = 1;
		fFormats[1].u.raw_video.display.format = B_RGBA32;

		fFormats_count = 2;
		fFlavors[0].in_format_count = 2;
		break;
	default:
		Log("VideoAddOn::VideoAddOn: supporting no colorspace at all!\n");
		/* this shouldn't happen.. */
		fFormats_count = 0;
		fFlavors[0].in_format_count = 0;
	}
}


VideoAddOn::~VideoAddOn()
{
	Log("VideoAddOn::~VideoAddOn\n");
	
	if (fStatus == B_OK)
	{
		SaveConfigurationFile();
		//BETA: We may need to "uninitialize" the video output hardare here...
	}
}


status_t VideoAddOn::InitCheck( const char ** outFailureText )
{
	Log("VideoAddOn::InitCheck\n");
	
	if (fStatus != B_OK)
		*outFailureText = "Hardware not found!";
	
	return fStatus;
}


int32 VideoAddOn::CountFlavors()
{
	Log("VideoAddOn::CountFlavors\n");
	
	if (fStatus != B_OK)
		return fStatus;
	
	return fFlavors_count;
}


status_t VideoAddOn::GetFlavorAt( int32 n, const flavor_info ** outInfo )
{
	Log("VideoAddOn::GetFlavorAt\n");
	
	if (fStatus != B_OK)
		return fStatus;
	
	if (n >= 0 && n < fFlavors_count)
	{
		*outInfo = &fFlavors[n];
		return B_OK;
	}
	
	return B_BAD_INDEX;
}


BMediaNode * VideoAddOn::InstantiateNodeFor (
	const flavor_info * info,
	BMessage * /* configuration */,
	status_t * outError )
{
	Log("VideoAddOn::InstantiateNodeFor\n");
	
	if (fStatus != B_OK)
	{
		*outError = B_NO_INIT;
		return NULL;
	}
	
	if (info->internal_id == fFlavors[0].internal_id)
	{
		VideoConsumer * node = fConsumer;
		fConsumer = new VideoConsumer(
			this, fFlavors[0].name, fFlavors[0].internal_id, iDrvInfo);
			
		if (fConsumer != NULL && fConsumer->InitCheck() != B_OK)
		{
			*outError = fConsumer->InitCheck();
			delete fConsumer;
			fConsumer = node;
			return NULL;
		}
		
		#if 1
			//HACK: load configuration settings
			LoadConfigurationFor(fConsumer);
		#endif
		
		return fConsumer;
	}
	
	return NULL;
}


status_t VideoAddOn::LoadConfigurationFile()
{
	Log("VideoAddOn::LoadConfigurationFile\n");
	
	BFile file(kSettingsFileName, B_READ_ONLY);
	if (file.InitCheck() == B_OK)
		return fConfigurations.Unflatten(&file);
	
	return B_ERROR;
}


status_t VideoAddOn::SaveConfigurationFile()
{
	Log("VideoAddOn::SaveConfigurationFile\n");
	
	BFile file(kSettingsFileName, B_CREATE_FILE | B_WRITE_ONLY);
	if (file.InitCheck() == B_OK)
		return fConfigurations.Flatten(&file);
	
	return B_ERROR;
}


status_t VideoAddOn::LoadConfigurationFor( BMediaNode * node )
{
	Log("VideoAddOn::LoadConfigurationFor\n");
	
	int32 internal_id;
	if (node != NULL && node->AddOn(&internal_id) == this)
	{
		if (node == fConsumer && internal_id == fFlavors[0].internal_id)
		{
			BMessage message;
			if (fConfigurations.FindMessage("VideoConsumer", &message) == B_OK)
				fConsumer->SetConfiguration(&message);
			return B_OK;
		}
	}
	return B_ERROR;
}


status_t VideoAddOn::SaveConfigurationFor( const BMediaNode * node )
{
	Log("VideoAddOn::SaveConfigurationFor\n");
	
	int32 internal_id;
	if (node != NULL && node->AddOn(&internal_id) == this)
	{
		if (node == fConsumer && internal_id == fFlavors[0].internal_id)
		{
			BMessage message;
			fConsumer->GetConfiguration(&message);
			if (fConfigurations.HasMessage("VideoConsumer"))
				fConfigurations.ReplaceMessage("VideoConsumer", &message);
			else
				fConfigurations.AddMessage("VideoConsumer", &message);
			return B_OK;
	  	}
	}
	return B_ERROR;
}


status_t VideoAddOn::GetConfigurationFor(
	BMediaNode * node,
	BMessage * configuration )
{
	Log("VideoAddOn::GetConfigurationFor\n");
	
	int32 internal_id;
	if (node->AddOn(&internal_id) == this)
	{
		if (node == fConsumer && internal_id == fFlavors[0].internal_id)
		{
			fConsumer->GetConfiguration(configuration);
			return B_OK;
		}
	}
	return B_ERROR;
}


bool VideoAddOn::WantsAutoStart()
{
	Log("VideoAddOn::WantsAutoStart\n");
	
	return false;
}


status_t VideoAddOn::AutoStart(
	int /* inCount */,
	BMediaNode ** /* outNode */,
	int32 * /*outInternalID */,
	bool * /* outHasMore */ )
{
	Log("VideoAddOn::AutoStart\n");
	
	return B_ERROR;
}


BMediaAddOn * make_media_addon( image_id image )
{
	return new VideoAddOn(image);
}


uint32 VideoAddOn::GetSelectedDevCaps(const BMessage *fConfigurations, const driver_info *iDrvInfo)
{
	//don't select the first device (0), as it _probably_ displays the Desktop:
	uint32 first_candidate = 1;
	int32 iParam_GraphicsDeviceID = 1;
	bool iConfigSet = false;
	BMessage message;

	Log ("VideoAddOn::GetSelectedDevCaps\n");

	//find the first device in case loading configfile failed or
	//was inconsistent with current hardware setup..
	for (uint32 idx(0); idx < 32; idx++)
	{
		if (iDrvInfo->ok_and_free_acc_at[idx].present)
		{
			first_candidate = idx;
			break;
		}
	}

	//get the current config (as was saved) if possible..
	if (fConfigurations->FindMessage("VideoConsumer", &message) == B_OK)
	{
		if (message.FindInt32("DeviceID", &iParam_GraphicsDeviceID) != B_OK)
		{
			Log("VideoAddOn::SetConfiguration failed!\n");
			// the first device _probably_ has the desktop, so don't select that
			iParam_GraphicsDeviceID = 1;
		}
		else
		{
			iConfigSet = true;
		}
	}

	//if loading configfile failed, or it gave inconsistant results,
	//select the first supported device ourself
	if (!iConfigSet || !(iDrvInfo->ok_and_free_acc_at[iParam_GraphicsDeviceID].present))
	{
		Log ("VideoAddOn::GetSelectedDevCaps: using first supported device found\n");
		iParam_GraphicsDeviceID = first_candidate;
	}
	Log ("VideoAddOn::GetSelectedDevCaps: iParam_GraphicsDeviceID is %d\n",
		iParam_GraphicsDeviceID);

	return iDrvInfo->ok_and_free_acc_at[iParam_GraphicsDeviceID].drw_modes;
}
