// VideoConsumer.cpp
//VGA Device Output consumer node, 18 april 2006, alpha 3 release.

#include <string.h>
#include <Buffer.h>
#include <BufferGroup.h>
#include <TimeSource.h>
#include <ParameterWeb.h>

#include "VideoAddOn.h"
#include "VideoConsumer.h"
#include "Log.h"
#include <iostream.h>

vint32 VideoConsumer::iInstances = 0;

VideoConsumer::VideoConsumer(
	BMediaAddOn * uAddOn,
	const char * uName,
	int32 uInternalID,
	driver_info eDrvInfo )
:
	BMediaNode(uName),
	BMediaEventLooper(),
	BBufferConsumer(B_MEDIA_RAW_VIDEO),
	BControllable(),
	
	iInitCheck(B_NO_INIT),
	
	//configuration has not been loaded yet
	iConfigSet(false),

	iUsingBlit (false),
	
	iAddOn(uAddOn),
	iInternalID(uInternalID),
	
	iBufferGroupBlit(NULL),
	iBufferGroupOverlay(NULL),
	iNumBufs (1),
	iPrefBuf (NULL),
	
	iLatency(0LL),
	
	iInput(),
	iPreferredFormat(),
	
	iDeviceManager(new DeviceManager),
	iGraphicsDevice(NULL),
	
	// the first device _probably_ has the desktop, so don't select that
	iParam_GraphicsDeviceID(1),
	
	iParamTime_GraphicsDeviceID(0LL),
	
	iOutputFramePeriod(0)
{
	Log("VideoConsumer::VideoConsumer\n");

	//copy the already setup free devices info for our use..
	iDrvInfo = eDrvInfo;
 
	if (atomic_add(&iInstances, 1) == 0)
	{
		VideoAddOn::fVideoConsumerPort = ControlPort();

		//Initialize our preferred format
		iPreferredFormat.type = B_MEDIA_RAW_VIDEO;
		iPreferredFormat.u.raw_video = media_raw_video_format::wildcard;
		iPreferredFormat.u.raw_video.interlace = 1;
		iPreferredFormat.u.raw_video.orientation = B_VIDEO_TOP_LEFT_RIGHT;

		/* initialize input */
		iInput.source = media_source::null;
		iInput.format = iPreferredFormat;
		
		AddNodeKind(B_PHYSICAL_OUTPUT);
		
		iInitCheck = B_OK;
	}
}


VideoConsumer::~VideoConsumer()
{
	Log("VideoConsumer::~VideoConsumer\n");
	if (iInitCheck == B_OK)
	{
		//HACK: Save configuration settings
		if (iAddOn != NULL)
			((VideoAddOn *) iAddOn)->SaveConfigurationFor(this);
		VideoAddOn::fVideoConsumerPort = 0;

		if (iBufferGroupBlit != NULL)
		{
			//get all buffers back..
			iBufferGroupBlit->ReclaimAllBuffers();
			iBufferGroupBlit->WaitForBuffers();
			//and Release buffer group
			iGraphicsDevice->DeleteBlitBufferGroup( iBufferGroupBlit );
			//must be done HERE, doesn't work if done in iGraphicsDevice!!!
			iBufferGroupBlit = NULL;
			Log("VideoConsumer::~VideoConsumer: Old blit buffergroup deleted\n");
		}
		if (iBufferGroupOverlay != NULL)
		{
			//get all buffers back..
			iBufferGroupOverlay->ReclaimAllBuffers();
			iBufferGroupOverlay->WaitForBuffers();
			//and Release buffer group
			iGraphicsDevice->DeleteOverlayBufferGroup( iBufferGroupOverlay );
			//must be done HERE, doesn't work if done in iGraphicsDevice!!!
			iBufferGroupOverlay = NULL;
			Log("VideoConsumer::~VideoConsumer: Old overlay buffergroup deleted\n");
		}
	}

	delete iGraphicsDevice;
	delete iDeviceManager;
	atomic_add(&iInstances, -1);

	Quit();
}


status_t VideoConsumer::InitCheck() const
{
	return iInitCheck;
}


status_t VideoConsumer::GetConfiguration( BMessage * outMessage ) const
{
	Log("VideoConsumer::GetConfiguration\n");
	
	outMessage->AddInt32("DeviceID", iParam_GraphicsDeviceID);
	
	return B_OK;
}


status_t VideoConsumer::SetConfiguration( BMessage * inMessage )
{
	Log("VideoConsumer::SetConfiguration\n");
	
	if (inMessage->FindInt32("DeviceID", &iParam_GraphicsDeviceID) != B_OK)
	{
		Log("VideoConsumer::SetConfiguration failed!\n");
		// the first device _probably_ has the desktop, so don't select that
		iParam_GraphicsDeviceID = 1;
	}
	else
	{
		iConfigSet = true;
	}

	return B_OK;
}


// BMediaNode

port_id VideoConsumer::ControlPort() const
{
	return BMediaNode::ControlPort();
}


BMediaAddOn * VideoConsumer::AddOn( int32 * cookie ) const
{
	*cookie = iInternalID;
	return iAddOn;
}


status_t VideoConsumer::HandleMessage(int32 message, const void * data,
	size_t size )
{
	Log("VideoConsumer::HandleMessage\n");
	// We don't use any private message protocols
	return BBufferConsumer::HandleMessage( message, data, size );
}


status_t VideoConsumer::RequestCompleted( const media_request_info & /* info */ )
{
	 Log("VideoConsumer::RequestCompleted\n");
	 return B_OK;
}


// BMediaEventLooper

void VideoConsumer::NodeRegistered() {
	Log("VideoConsumer::NodeRegistered\n");
	
	if (iInitCheck != B_OK) {
		ReportError( B_NODE_IN_DISTRESS );
	}	else {
		iInput.node = Node();
		iInput.destination.port = ControlPort();
		iInput.destination.id = 0;
		iInput.source = media_source::null;
		strncpy( iInput.name, Name(), B_MEDIA_NAME_LENGTH );
		
		SetParameterWeb( CreateParameterWeb() );
		//CreateParameterWeb as called above inits our output device, so:
		if (iGraphicsDevice)
		{
			//fixme: ATI supports RGB overlay(?)
			if (iGraphicsDevice->GetDrawModes() & VID_OVERLAY)
			{
				Log("VideoConsumer::NodeRegistered: Prefering OVERLAY\n");
				iPreferredFormat.u.raw_video.display.format = B_YCbCr422;
			}
			else
			{
				Log("VideoConsumer::NodeRegistered: Prefering SCALED_FILTERED_BLIT\n");
				iPreferredFormat.u.raw_video.display.format = B_RGB32;
			}
		}
		
		iInput.format = iPreferredFormat;
		
		BMediaEventLooper::NodeRegistered();
	}
}


void VideoConsumer::Start( bigtime_t performance_time )
{
	BMediaEventLooper::Start( performance_time );
}


void VideoConsumer::Stop( bigtime_t performance_time, bool immediate)
{
	BMediaEventLooper::Stop( performance_time, immediate );
}


void VideoConsumer::Seek( bigtime_t media_time, bigtime_t performance_time )
{
	BMediaEventLooper::Seek( media_time, performance_time );
}


void VideoConsumer::TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time )
{
	BMediaEventLooper::TimeWarp( at_real_time, to_performance_time );
}


void VideoConsumer::HandleEvent(const media_timed_event *Event,
	bigtime_t /* Lateness */, bool /* RealTimeEvent */ )
{
	switch (Event->type)
	{
	case BTimedEventQueue::B_HANDLE_BUFFER:
		{
			BBuffer* Buffer = static_cast<BBuffer *>( Event->pointer );

			if (Buffer != NULL)
			{
				if (RunState() == BMediaEventLooper::B_STARTED)
				{
					//drop late frames.. (otherwise we'd loose audio/video sync)
					if (Event->event_time >= TimeSource()->Now())
					{
						if (iInput.source != media_source::null)
						{
							if (iUsingBlit)
							{
								iGraphicsDevice->AcquireEngine();
								iGraphicsDevice->WaitForRetrace();
								iGraphicsDevice->ScaledBlit(Buffer->Flags());
								//buffer will be overwritten if there's just one...
								if (iNumBufs == 1)
									iGraphicsDevice->WaitEngineIdle();
								iGraphicsDevice->ReleaseEngine();
							}
							else
							{
								iGraphicsDevice->ConfigureOverlay(Buffer->Flags());
								//notes:
								//- single buffered overlay might _seem_ not
								//  synced to retrace although it is: this is normal
								//  and the effect is caused by the fact that although
								//  we recycle the buffer back to the BBufferProducer
								//  we keep using it!!
								//  (producer copies new data in it unsynced to retrace.)
								//
								//- when single buffering overlay we need to call
								//  ConfigureOverlay() just _once_.
								//  Being lazy here for now..
							}
						}
					}
					else Log("=Frame Dropped=\n");
				}
				if (iNumBufs == 1)
				{
					Buffer->Recycle();
				}
				else
				{
					//if we use multiple buffers we can keep one 'busy'...
					if (iPrefBuf)
						iPrefBuf->Recycle();
					iPrefBuf = Buffer;
				}
			}
			else Log("ERROR: NULL Buffer Received?\n");
		}
		break;

	case BTimedEventQueue::B_START:
		Log("VideoConsumer::HandleEvent:Start\n");
		if (RunState() != BMediaEventLooper::B_STARTED) {
			//BETA: Start the video?
		}
		break;
		
	case BTimedEventQueue::B_STOP:
		Log("VideoConsumer::HandleEvent:Stop\n");
		if (RunState() == BMediaEventLooper::B_STARTED) {
			//BETA: Stop the video stream?
			EventQueue()->FlushEvents( 0,
				BTimedEventQueue::B_ALWAYS, true,
				BTimedEventQueue::B_HANDLE_BUFFER );
		}
		break;
		
	case BTimedEventQueue::B_SEEK:
		Log("VideoConsumer::HandleEvent:Seek\n");
		break;
		
	case BTimedEventQueue::B_WARP:
		Log("VideoConsumer::HandleEvent:Warp\n");
		break;
		
	case BTimedEventQueue::B_PARAMETER:
		Log("VideoConsumer::HandleEvent:Parameter\n");
		break;
		
	}
}


status_t VideoConsumer::DeleteHook( BMediaNode * node )
{
	return BMediaEventLooper::DeleteHook(node);
}


/* BBufferConsumer */

status_t VideoConsumer::AcceptFormat(
	const media_destination & destination,
	media_format * format )
{
	Log("VideoConsumer::AcceptFormat\n");

	if (iInput.destination != destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (format->type != B_MEDIA_RAW_VIDEO)
		return B_MEDIA_BAD_FORMAT;
	
	return SpecializeFormat(format);
}


status_t VideoConsumer::GetNextInput( int32 * cookie, media_input * outInput )
{
	Log("VideoConsumer::GetNextInput\n");
	
	if (*cookie != 0)
		return B_BAD_INDEX;
	
	*outInput = iInput;
	*cookie += 1;
	
	return B_OK;
}


void VideoConsumer::DisposeInputCookie( int32 /* cookie */ )
{
	Log("VideoConsumer::DisposeInputCookie\n");
}


void VideoConsumer::BufferReceived( BBuffer * Buffer )
{
	if (Buffer->Header()->type == B_MEDIA_PARAMETERS)
	{
		//It's a special Buffer with parameter data
		ApplyParameterData( Buffer->Data(), Buffer->SizeUsed() );
		Buffer->Recycle();
	}
	else
	{
	  	bigtime_t EventTime( Buffer->Header()->start_time );

		//It's a normal buffer: Try adding it to our internal queue.
		if (EventQueue()->AddEvent( media_timed_event(
				EventTime,
				BTimedEventQueue::B_HANDLE_BUFFER,
				Buffer,
				BTimedEventQueue::B_RECYCLE_BUFFER) ) < B_OK )
		{
			//adding to queue failed, so we must recycle this buffer without using!
			Buffer->Recycle();
		}
	}
}


void VideoConsumer::ProducerDataStatus(
	const media_destination &for_whom,
	int32 status,
	bigtime_t at_performance_time )
{
	Log("VideoConsumer::ProducerDataStatus\n");

	//HACK: Added the following code from the BeBook
	//Is the HandleEvent hook doing anything with these?
	if (for_whom == iInput.destination)
	{
		media_timed_event event(
			at_performance_time,
			BTimedEventQueue::B_DATA_STATUS, &iInput,
			BTimedEventQueue::B_NO_CLEANUP, status, 0, NULL );
		EventQueue()->AddEvent( event );
	}
}


status_t VideoConsumer::GetLatencyFor(
	const media_destination &for_whom,
	bigtime_t * outLatency,
	media_node_id * outTimeSource )
{
	Log("VideoConsumer::GetLatencyFor\n");
	
	if (for_whom != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;
	
	*outLatency = iLatency;
	*outTimeSource = TimeSource()->ID();
	
	return B_OK;
}


status_t VideoConsumer::Connected(
	const media_source & source,
	const media_destination & destination,
	const media_format & format,
	media_input * outInput )
{
	Log("VideoConsumer::Connected\n");
	
	if (destination != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;

	if (iGraphicsDevice != NULL)
	{
		//delete buffers used on old format..
		if (iBufferGroupBlit != NULL)
		{
			//get all buffers back..
			iBufferGroupBlit->ReclaimAllBuffers();
			iBufferGroupBlit->WaitForBuffers();
			//and Release buffer group
			iGraphicsDevice->DeleteBlitBufferGroup( iBufferGroupBlit );
			//must be done HERE, doesn't work if done in iGraphicsDevice!!!
			iBufferGroupBlit = NULL;
			Log("VideoConsumer::~VideoConsumer: Old blit buffergroup deleted\n");
		}
		if (iBufferGroupOverlay != NULL)
		{
			//get all buffers back..
			iBufferGroupOverlay->ReclaimAllBuffers();
			iBufferGroupOverlay->WaitForBuffers();
			//and Release buffer group
			iGraphicsDevice->DeleteOverlayBufferGroup( iBufferGroupOverlay );
			//must be done HERE, doesn't work if done in iGraphicsDevice!!!
			iBufferGroupOverlay = NULL;
			Log("VideoConsumer::~VideoConsumer: Old overlay buffergroup deleted\n");
		}

		//create buffers for the new format...
		if (iUsingBlit)
		{
			//fixme: use a custom version as soon as the driver has new hooks!
			if (!iBufferGroupBlit)
				iBufferGroupBlit = iGraphicsDevice->CreateFixedBlitBufferGroup(&iNumBufs);

			//setup blitlist for the current format
			iGraphicsDevice->SetScaledBlitList(iNumBufs, format);

			//Specify the buffer group for the connection
			int32 userData = 0, changeTag = 0;
		
			if (SetOutputBuffersFor(
				source, destination, iBufferGroupBlit,
			  (void *) &userData, &changeTag, true ) == B_OK)
			  Log("New (blit) buffers have been set!\n");
		}
		else
		{
			if (!iBufferGroupOverlay)
				iBufferGroupOverlay = iGraphicsDevice->CreateOverlayBufferGroup(&iNumBufs, format);

			//setup overlay zoomfactor for the current format
			iGraphicsDevice->SetOverlayZoom(iNumBufs, format);

			//acquire overlay engine
			iGraphicsDevice->AllocateOverlay();

			//Specify the buffer group for the connection
			int32 userData = 0, changeTag = 0;
		
			if (SetOutputBuffersFor(
				source, destination, iBufferGroupOverlay,
			  (void *) &userData, &changeTag, true ) == B_OK)
			  Log("New (overlay) buffers have been set!\n");
		}
		
		//for doublebuffered output there's no previous sent frame yet
		iPrefBuf = NULL;

		//erase rubbish from screen, so blackness is displayed
		iGraphicsDevice->ClearScreen();

		//turn-on the screen
		iGraphicsDevice->SetDPMS(B_DPMS_ON);

		// set the buffers duration in uS (time it's visible on-screen)
		SetBufferDuration(1000000LL / format.u.raw_video.field_rate );

		//Set the node's latency in uS (delay until shown on-screen)
		//note: this depends on the screen's refresh rate
		iOutputFramePeriod = bigtime_t( 1000000.0f / 70.0f );
		iLatency = iOutputFramePeriod + (iOutputFramePeriod >> 1);
		SetEventLatency( iLatency );

		//Reserve the connection
		iInput.source = source;
		iInput.format = format;

		*outInput = iInput;
		
		return B_OK;
	}
	
	return B_ERROR;
}


void VideoConsumer::Disconnected(
	const media_source & producer,
	const media_destination & where )
{
	Log("VideoConsumer::Disconnected\n");

	if (where == iInput.destination && producer == iInput.source)
	{
		//switch off the screen
		iGraphicsDevice->SetDPMS(B_DPMS_STAND_BY);

		if (iUsingBlit)
		{
			//first recycle all buffers..
			int32 bufs;
			BBuffer* buffers[3];
			if ((iBufferGroupBlit->CountBuffers(&bufs) == B_OK) &&
				(iBufferGroupBlit->GetBufferList(bufs, &(buffers[0])) == B_OK))
			{
				Log("VideoConsumer::recycling buffers\n");
				for (int32 cnt = 0; cnt < bufs; cnt++)
				{
					buffers[cnt]->Recycle();
				}
			}
		}
		else
		{
			//release overlay engine
			iGraphicsDevice->ReleaseOverlay();

			//now recycle all buffers..
			int32 bufs;
			BBuffer* buffers[3];
			if ((iBufferGroupOverlay->CountBuffers(&bufs) == B_OK) &&
				(iBufferGroupOverlay->GetBufferList(bufs, &(buffers[0])) == B_OK))
			{
				Log("VideoConsumer::recycling buffers\n");
				for (int32 cnt = 0; cnt < bufs; cnt++)
				{
					buffers[cnt]->Recycle();
				}
			}
		}

		//then remove the buffer group for the connection
		int32 userData = 0, changeTag = 0;
		//inform BMediaRoster that we will reclaim our buffers ourselves..
		SetOutputBuffersFor(
			producer, where, NULL,
			(void *) &userData, &changeTag, true );

		//note: don't reclaim or delete buffers here, but in Connected. (see BeBook)

		//Unreserve the connection
		iInput.source = media_source::null;
		iInput.format = iPreferredFormat;
	}
}


status_t VideoConsumer::FormatChanged(
	const media_source & producer,
	const media_destination & consumer,
	int32 /* from_change_count */,
	const media_format &/*format */)
{
	Log("VideoConsumer::FormatChanged\n");
	
	if (consumer != iInput.destination)
		return B_MEDIA_BAD_DESTINATION;
	
	if (producer != iInput.source)
		return B_MEDIA_BAD_SOURCE;
	
	return B_ERROR;
}


// BControllable

status_t VideoConsumer::GetParameterValue(
	int32 id,
	bigtime_t * last_change,
	void * value,
	size_t * size )
{
	switch (id)
	{
	case Param_DeviceID:
		*last_change = iParamTime_GraphicsDeviceID;
		*((int32 *) value) = iParam_GraphicsDeviceID;
		*size = sizeof(int32);
		break;
	}
	return B_OK;
}


void VideoConsumer::SetParameterValue(int32 id, bigtime_t when, const void * value,	size_t /* size */ )
{
	switch (id)
	{
	case Param_DeviceID:
		iParamTime_GraphicsDeviceID = when;
		iParam_GraphicsDeviceID = *((const int32 *) value);
		BroadcastNewParameterValue(when, id, &iParam_GraphicsDeviceID, sizeof(iParam_GraphicsDeviceID));
		break;
	}
	
	EventQueue()->AddEvent(media_timed_event(when,
		BTimedEventQueue::B_PARAMETER, NULL,
		BTimedEventQueue::B_NO_CLEANUP, id, 0, NULL, 0));
}


status_t VideoConsumer::StartControlPanel( BMessenger * /* messenger */ )
{
	return B_ERROR;
}


// Implementation

//(called by VideoConsumer::AcceptFormat())
status_t VideoConsumer::SpecializeFormat( media_format * format )
{
	if (format == NULL)	return B_BAD_VALUE;
	if (format->type != B_MEDIA_RAW_VIDEO) return B_MEDIA_BAD_FORMAT;
	if (!iGraphicsDevice) return B_ERROR;

	//check for a usable colorspace
	uint32 drw_modes = iGraphicsDevice->GetDrawModes();

	if (format->u.raw_video.display.format == media_raw_video_format::wildcard.display.format)
	{
		Log ("VideoConsumer::SpecializeFormat: Initial colorspace is wildcard\n"); 
	}
	else
	{
		Log ("VideoConsumer::SpecializeFormat: Initial colorspace is $%08x\n", 
			format->u.raw_video.display.format);
	}

	//Specialize raw video display fields
	//NOTE: Consumer nodes may only specialize items that are not already set!
	if (format->u.raw_video.display.format == media_raw_video_format::wildcard.display.format)
	{
		if (drw_modes & VID_OVERLAY)
			format->u.raw_video.display.format = B_YCbCr422;
		else
			format->u.raw_video.display.format = B_RGB32;
	}

	if (format->u.raw_video.display.format == B_YCbCr422)
	{
		if (drw_modes & VID_OVERLAY)
		{
			Log ("VideoConsumer::SpecializeFormat: using OVERLAY\n");
			iUsingBlit = false;
		}
		else
		{
			Log ("VideoConsumer::SpecializeFormat: colorspace unusable, aborting!\n");
			return B_MEDIA_BAD_FORMAT;
		}
	}
	//fixme: ATI driver can do RGB32 overlay(?)
	if ((format->u.raw_video.display.format == B_RGB32) ||
		(format->u.raw_video.display.format == B_RGBA32)) //3ivX decoder
	{
		if (drw_modes & VID_BLIT)
		{
			Log ("VideoConsumer::SpecializeFormat: using SCALED_BLIT\n");
			iUsingBlit = true;
		}
		else
		{
			Log ("VideoConsumer::SpecializeFormat: space unusable, aborting!\n");
			return B_MEDIA_BAD_FORMAT;
		}
	}

	if (format->u.raw_video.display.line_width == media_raw_video_format::wildcard.display.line_width)
		format->u.raw_video.display.line_width = 640;
	if (format->u.raw_video.display.line_count == media_raw_video_format::wildcard.display.line_count)
		format->u.raw_video.display.line_count = 480;

	//workspace blit scale-down costs even more RAM, some overlay units can't scale
	//down. Could both be checked for constraints, but I'm lazy.
	if ((format->u.raw_video.display.line_width > 640) ||
		(format->u.raw_video.display.line_count > 480))
	{
		Log ("VideoConsumer::SpecializeFormat: not scaling down (video > screen), aborting!\n");
		return B_MEDIA_BAD_FORMAT;
	}

	//NOTE 1: some format decoders do NOT allow setting bytes_per_row here (no wildcards), especially on R5!
	//NOTE 2:
	//we should return B_MEDIA_BAD_FORMAT if we are called with a preset
	//u.raw_video.bytes_per_row that is not aligned to our buffer options.
	//Not doing that yet because I want to demonstrate which decoders 'malfunction'
	//concerning this for now..
	if (format->u.raw_video.display.bytes_per_row != media_raw_video_format::wildcard.display.bytes_per_row)
		Log ("VideoConsumer::SpecializeFormat: bytes_per_row is fixed by decoder node!\n");

	if (iUsingBlit)
	{
		if (format->u.raw_video.display.bytes_per_row == media_raw_video_format::wildcard.display.bytes_per_row)
		{
			frame_buffer_config fbc;
			iGraphicsDevice->GetFrameBufferConfig(fbc);
			format->u.raw_video.display.bytes_per_row = fbc.bytes_per_row;
		}
	}
	else
	{
		//deleting overlay buffers so we can create a new one for sure...
		if (iBufferGroupOverlay != NULL)
		{
			//get all buffers back..
			iBufferGroupOverlay->ReclaimAllBuffers();
			iBufferGroupOverlay->WaitForBuffers();
			//and Release buffer group
			iGraphicsDevice->DeleteOverlayBufferGroup( iBufferGroupOverlay );
			//must be done HERE, doesn't work if done in iGraphicsDevice!!!
			iBufferGroupOverlay = NULL;
			Log("VideoConsumer::SpecializeFormat: Old overlay buffergroup deleted\n");
		}

		uint16 bpr;
		const overlay_buffer *mob = iGraphicsDevice->AllocateOverlayBuffer(
			format->u.raw_video.display.line_width,
			format->u.raw_video.display.line_count);
		if (!mob)
		{
			Log ("VideoConsumer::SpecializeFormat: overlay buffer alloc failed, aborting!\n");
			return B_MEDIA_BAD_FORMAT;
		}
		iGraphicsDevice->GetOverlayBufferConfig(mob, &bpr);
		if (format->u.raw_video.display.bytes_per_row != media_raw_video_format::wildcard.display.bytes_per_row)
			Log ("VideoConsumer::SpecializeFormat: Forcing bytes_per_row!\n");
		//may officially only be done if it was set to wildcard (but for overlay setup this can't be specified).
		format->u.raw_video.display.bytes_per_row = bpr;
		iGraphicsDevice->ReleaseOverlayBuffer(mob);
	}
	Log ("VideoConsumer::SpecializeFormat: bytes_per_row is %d\n",
		format->u.raw_video.display.bytes_per_row);

	//Specialize raw video fields
	if (format->u.raw_video.field_rate == media_raw_video_format::wildcard.field_rate)
		format->u.raw_video.field_rate = 70.;
	if (format->u.raw_video.interlace == media_raw_video_format::wildcard.interlace)
	    format->u.raw_video.interlace = 1;
	if (format->u.raw_video.first_active == media_raw_video_format::wildcard.first_active)
		format->u.raw_video.first_active = 0;
	if (format->u.raw_video.last_active == media_raw_video_format::wildcard.last_active)
		format->u.raw_video.last_active = 479;
	if (format->u.raw_video.orientation == media_raw_video_format::wildcard.orientation)
		format->u.raw_video.orientation = B_VIDEO_TOP_LEFT_RIGHT;

	//Clip the first and last active lines (officially not allowed(?))
	if (format->u.raw_video.first_active >= format->u.raw_video.display.line_count)
		format->u.raw_video.first_active = format->u.raw_video.display.line_count - 1;
	if (format->u.raw_video.last_active >= format->u.raw_video.display.line_count)
		format->u.raw_video.last_active = format->u.raw_video.display.line_count - 1;
	
	//Clip the field rate  (officially not allowed(?))
	if (format->u.raw_video.field_rate <= 1.0f)
	    format->u.raw_video.field_rate = 1.0f;
	if (format->u.raw_video.field_rate >= 240.0f)
	    format->u.raw_video.field_rate = 240.0f;

	//If decoder asks for B_MEDIA_RETAINED_DATA then it _needs_ one buffer precicely.
	//(it uses old content of this buffer to create new content in same buffer (?))
	//This fixes (at least) cinepak.decoder multiple buffer decoding artifacts.
	if (format->require_flags & B_MEDIA_RETAINED_DATA)
	{
		Log("VideoConsumer::SpecializeFormat: Decoder implicitly wants one BBuffer\n");
		iNumBufs = 1;
	}
	else
	{
		Log("VideoConsumer::SpecializeFormat: Decoder can use multiple BBuffers\n");
		iNumBufs = 3;
	}

	return B_OK;
}

BParameterWeb * VideoConsumer::CreateParameterWeb()
{
	//don't select the first device (0), as it _probably_ displays the Desktop:
	uint32 first_candidate = 1;
	bool first_candidate_selected = false;

	BParameterWeb * web = new BParameterWeb();
	BParameterGroup * options1 = web->MakeGroup("Device");
	BDiscreteParameter * DParam(NULL);
	
	DParam = options1->MakeDiscreteParameter(Param_DeviceID,
		B_MEDIA_NO_TYPE, "Device", B_OUTPUT_MUX);

	//I just want to add some text here. Is this how it should be? (beginner.. :)
	options1->MakeNullParameter(Param_TextOnly,
		B_MEDIA_NO_TYPE, "Restart the Media Server to apply changes.", B_GENERIC);

	//iterate through all available graphics devices and get their names
	//we are already sure to have at least one useable and free device here!
	for (uint32 idx(0); idx < 32; idx++)
	{
		BString DeviceName;
		if (iDrvInfo.ok_and_free_acc_at[idx].present)
		{
			//populate 'Device' parameter:
			//note: some drivers seem to not set a name while still having
			//this function.... (?)
			//(although the Be unified Nvidia driver is just 'faulty':
			// that one gives back correct info after the _second_ or later time
			// we call the get_accelerant_device_info hook!)
			if (strlen(iDrvInfo.ok_and_free_acc_at[idx].name) == 0)
				DeviceName = BString("Unknown VGA Device");
			else
				DeviceName = iDrvInfo.ok_and_free_acc_at[idx].name;
			DParam->AddItem(idx, DeviceName.String());

			//find the first device in case loading configfile failed or
			//was inconsistent with current hardware setup..
			if (!first_candidate_selected)
			{
				first_candidate = idx;
				first_candidate_selected = true;
			}
		}
	}

	//if loading configfile failed, or it gave inconsistant results,
	//select the first supported device ourself
	if (!iConfigSet || !(iDrvInfo.ok_and_free_acc_at[iParam_GraphicsDeviceID].present))
	{
		Log ("VideoConsumer::CreateParameterWeb: using first supported device found\n");
		iParam_GraphicsDeviceID = first_candidate;
	}
	Log ("VideoConsumer::CreateParameterWeb: iParam_GraphicsDeviceID is %d\n",
		iParam_GraphicsDeviceID);

	//done here instead of in 'connected'...
	//get and init our graphics device: a node restart is needed to use another device:
	//which is done on purpose, as constantly (re)starting drivers costs much time...
	if (iDeviceManager == NULL)
			Log ("VideoConsumer::CreateParameterWeb: iDeviceManager is NULL!\n");
	if (iGraphicsDevice != NULL)
			Log ("VideoConsumer::CreateParameterWeb: iGraphicsDevice isn't NULL!\n");

	if ((iDeviceManager != NULL) && (iGraphicsDevice == NULL))
	{
		iGraphicsDevice = iDeviceManager->GetGraphicsDevice( iParam_GraphicsDeviceID );
	}

	return web;
}
